import torch
import torch as torch
import torch.nn as nn
import torch.nn.functional as F

import torch.optim as optim
from torch.optim import lr_scheduler

import numpy as np
import torchvision
from torchvision import datasets, models, transforms

import matplotlib.pyplot as plt
import time
import os
import copy

from torchsummary import summary
import visdom  # http://localhost:8097/

# plt.ion()   # interactive mode


import torch.onnx
import onnx
import onnxruntime

NUM_OF_CLASS = 2  # 이진 분류

# INPUT_DIMENSION = 3     # RGB
INPUT_DIMENSION = 1  # Grayscale

# For ImageMap
INPUT_H = 36
INPUT_W = 208

# For Scatterplot
# INPUT_H = 256
# INPUT_W = 208


# For FullGraph
# INPUT_H = 208
# INPUT_W = 2048


NUM_OF_EPOCHS = 1000

SIZE_OF_BATCH = 4
# SIZE_OF_BATCH = 2


# Visdom 참고: https://keep-steady.tistory.com/12
# 프로그램 run 전에 터미널에서 명령 실행 : "python -m visdom.server"
# visdom -> http://localhost:8097/
USE_VISDOM: bool = True

# -----------------------------------------------------------------------------------------------------------------------
# Training 및 Test 데이터셋 로드 / 정규화
# -----------------------------------------------------------------------------------------------------------------------
# 학습을 위한 데이터 증가(Augmentation)와 일반화하기
# 단지 검증을 위한 일반화하기
data_transforms = {
    'train': transforms.Compose([
        # transforms.RandomResizedCrop(224),
        # transforms.RandomHorizontalFlip(),
        # transforms.Resize(256),
        # transforms.CenterCrop(224),
        transforms.Resize((INPUT_H, INPUT_W)),
        transforms.Grayscale(num_output_channels=INPUT_DIMENSION),  # transforms.Normalize(mean=[0.5], std=[0.5]),
        transforms.ToTensor(),
        # transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        # transforms.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])     # 데이터 로드 시 3채널(R,G,B) 데이터의 평균과 분산을 0.5로 설정
    ]),
    'val': transforms.Compose([
        # transforms.Resize(256),
        # transforms.CenterCrop(224),
        transforms.Resize((INPUT_H, INPUT_W)),
        transforms.Grayscale(num_output_channels=INPUT_DIMENSION),  # transforms.Normalize(mean=[0.5], std=[0.5]),
        transforms.ToTensor(),
        # transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        # transforms.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])     # 데이터 로드 시 3채널(R,G,B) 데이터의 평균과 분산을 0.5로 설정
    ]),
}

# data_dir = 'banknote_data_Scatterplot'
# data_dir = 'banknote_data_FullGraph'
data_dir = 'banknote_data_ImageMap'

path = {x: os.path.join(os.path.dirname(os.path.abspath(__file__)), data_dir, x)
        for x in ['train', 'val']}

# path['train']은 train set의 경로
# path['val']은 val set의 경로
# join은 문자열을 이어 붙여주는 함수

image_datasets = {x: datasets.ImageFolder(path[x], data_transforms[x])
                  for x in ['train', 'val']}

# num_workers 가 0보다 크면 오류 발생 ??
dataloaders = {'train': torch.utils.data.DataLoader(image_datasets['train'],
                                                    SIZE_OF_BATCH, shuffle=True, num_workers=0),

               'val': torch.utils.data.DataLoader(image_datasets['val'],
                                                  SIZE_OF_BATCH, shuffle=True, num_workers=0)}

dataset_sizes = {x: len(image_datasets[x]) for x in ['train', 'val']}
# dataset_sizes['train'] = train set 사진 갯수
# dataset_sizes['val'] = val set 사진 갯수

# class_names = image_datasets['train'].classes
# class_names = ['ants', 'bees']

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
print(device)  # GPU가 이용 가능한지 확인


# -----------------------------------------------------------------------------------------------------------------------


# -----------------------------------------------------------------------------------------------------------------------
# 일부 이미지 시각화
# -----------------------------------------------------------------------------------------------------------------------
def imshow(inp, title=None):
    """Imshow for Tensor."""
    inp = inp.numpy().transpose((1, 2, 0))
    # mean = np.array([0.485, 0.456, 0.406])
    # std = np.array([0.229, 0.224, 0.225])
    # mean = np.array([0.5, 0.5, 0.5])
    # std = np.array([0.5, 0.5, 0.5])
    # mean = np.array([0.5])
    # std = np.array([0.5])
    # inp = std * inp + mean   # 정규화를 해제
    inp = np.clip(inp, 0, 1)
    plt.imshow(inp)
    if title is not None:
        plt.title(title)
    plt.pause(1)


# 한개의 배치(batch)만큼 이미지를 불러온다. 배치 사이즈를 4로 했으니 사진 4장이 로드된다.
# inputs, classes = next(iter(dataloaders['train']))

# 로드된 데이터에 make_grid 함수를 통해 그리드를 추가한다.
# out = torchvision.utils.make_grid(inputs)

# 이미지를 출력한다.
# imshow(out, title=[class_names[x] for x in classes])


# -----------------------------------------------------------------------------------------------------------------------


# -----------------------------------------------------------------------------------------------------------------------
# LeNet-5 참고 :
# https://tutorials.pytorch.kr/beginner/blitz/neural_networks_tutorial.html
# -----------------------------------------------------------------------------------------------------------------------
class LeNet5(nn.Module):
    def __init__(self):
        super(LeNet5, self).__init__()
        # 1 input image channel, 6 output channels, 3x3 square convolution
        # kernel
        self.conv1 = nn.Conv2d(in_channels=INPUT_DIMENSION, out_channels=6, kernel_size=3)
        self.conv2 = nn.Conv2d(in_channels=6, out_channels=16, kernel_size=3)

        # an affine operation: y = Wx + b
        # self.fc1 = nn.Linear(16 * 6 * 6, 120)  # 6*6 from image dimension
        self.fc1 = nn.Linear(16 * 49 * 6, 120)  # 208 * 36 인 경우 (ImageMap)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, NUM_OF_CLASS)  # 2 = 이진분류

    def forward(self, x):  # [SIZE_OF_BATCH, INPUT_DIMENSION, INPUT_W, INPUT_H]
        out = F.relu(self.conv1(x))  # [4, 1, 208, 36] -> [4, 6, 204, 32]
        out = F.max_pool2d(out, kernel_size=2, stride=2)  # [4, 6, 204, 32] -> [4, 6, 102, 16]
        out = F.relu(self.conv2(out))  # [4, 6, 102, 16] -> [4, 16, 98, 12]
        out = F.max_pool2d(out, kernel_size=2, stride=2)  # [4, 16, 98, 12] -> [4, 16, 49, 6] -> 입력 : (16 * 49 * 6)
        out = out.view(out.size(0), -1)
        out = F.relu(self.fc1(out))
        out = F.relu(self.fc2(out))
        out = self.fc3(out)
        return out


class KisanLeNet5(nn.Module):
    def __init__(self):
        super(KisanLeNet5, self).__init__()

        # TIDL Limitation : Conv Layer 의 커널 크기는 7x7 이 최대, Stride 는 1 or 2, In/Out Channel 은 1024 가 최대
        self.conv1 = nn.Conv2d(in_channels=INPUT_DIMENSION, out_channels=6, kernel_size=5)
        self.conv2 = nn.Conv2d(in_channels=6, out_channels=12, kernel_size=5)
        self.conv2_bn = nn.BatchNorm2d(12)

        # self.fc1 = nn.Linear(16 * 49 * 61, 120)  # 208 * 256 인 경우 (Scatterplot)
        # self.fc1 = nn.Linear(16 * 509 * 49, 120)  # 4096 * 208 인 경우 (FullGraph)

        # TIDL Limitation : FC Layer 의 Input 은 4096 이 넘으면 안됨, 8의 배수여야 함
        self.fc1 = nn.Linear(12 * 49 * 6, 120)  # 208 * 36 인 경우 (ImageMap)
        self.fc1_bn = nn.BatchNorm1d(120)
        self.fc2 = nn.Linear(120, 40)
        self.fc2_bn = nn.BatchNorm1d(40)
        self.fc3 = nn.Linear(40, NUM_OF_CLASS)  # 2 = 이진분류

    def forward(self, x):  # [SIZE_OF_BATCH, INPUT_DIMENSION, INPUT_W, INPUT_H]
        out = F.relu(self.conv1(x))  # [4, 1, 208, 36] -> [4, 16, 204, 32]
        out = F.max_pool2d(out, kernel_size=2, stride=2)  # [4, 16, 204, 32] -> [4, 16, 102, 16]
        out = F.relu(self.conv2_bn(self.conv2(out)))
        #out = F.relu(self.conv2(out))                     # [4, 16, 102, 16] -> [4, 32, 98, 12]
        out = F.max_pool2d(out, kernel_size=2, stride=2)  # [4, 32, 98, 12] -> [4, 32, 49, 6] -> 입력 : (32 * 49 * 6)

        # TIDL Limitation : Reshape 레이어 대신 Flatten 레이어로 변환하여 입력해야 함
        # out = out.view(out.size(0), -1)
        out = torch.flatten(out, start_dim=1, end_dim=-1)  # https://pytorch.org/docs/master/torch.html#torch.flatten
        out = F.relu(self.fc1_bn(self.fc1(out)))
        #out = F.relu(self.fc1(out))
        out = F.relu(self.fc2_bn(self.fc2(out)))
        #out = F.relu(self.fc2(out))
        out = self.fc3(out)
        return out


# -----------------------------------------------------------------------------------------------------------------------

# -----------------------------------------------------------------------------------------------------------------------
# 모델 정의 (Pre-trained model 이용)
# 미리 학습된 모델의 가중치를 갖고와서 완전 연결 계층(Fully Connected Layer)을 학습시켜 2 Class 이미지 분류 진행
# -----------------------------------------------------------------------------------------------------------------------
# model_ft = models.alexnet(pretrained=False)
# num_ftrs = 4096
# device = "cpu"

model_ft = KisanLeNet5()
# model_ft = models.resnet18(pretrained=False)     # 미리 훈련된 모델의 weight 를 가지고 오도록 함 -> Transfer Learning (전이 학습)
# num_ftrs = model_ft.fc.in_features
# model_ft.fc = nn.Linear(num_ftrs, 2).cuda()     # 노드 수 : 2 (이진 분류)
model_ft = model_ft.to(device)

print(model_ft)
summary(model_ft, input_size=(INPUT_DIMENSION, INPUT_H, INPUT_W))

# -----------------------------------------------------------------------------------------------------------------------


# -----------------------------------------------------------------------------------------------------------------------
# 손실 함수(Loss Function)와 Optimizer 정의
# -----------------------------------------------------------------------------------------------------------------------
criterion = nn.CrossEntropyLoss()  # Loss 함수로 CrossEntropy 사용

# Observe that all parameters are being optimized
# optimizer_ft = optim.SGD(model_ft.parameters(), lr=0.001,
#                         momentum=0.9)  # Optimizer 로 Stochastic Gradient Descent(SGD) 사용
optimizer_ft = optim.Adam(model_ft.parameters())  # Optimizer 로 Adam 알고리즘 사용
# lr = learning rate

# SGD의 경우, 기존의 Gradient Descent와 비교해서 크게 두 가지의 장점이 있다.
# 첫 번째로는 GD를 사용하는 경우, weight update를 진행 할 때 전체 데이터에 대해 Loss 함수를 계산해야 하지만,
# SGD의 경우 전체 데이터 대신 일부 데이터의 모음(mini batch)에 대해서만 Loss 함수를 계산한다.
# 두 번째로는 SGD의 경우 최적점으로 이동하는 과정에서 Momentum이라는 성질을 이용하게 된다.
# 앞서 말했던 것처럼, SGD는 일부 데이터의 모음(mini batch)에 대해서만
# loss 함수를 계산하고 그 기울기를 weight update에 사용하기 때문에
# Learning rate가 작다 하더라도 update되는 weight가 매우 noisy하다.
# 따라서 Momentum이라는 개념을 추가하여, update 되는 weight의 비율을 상대적으로 더 줄임으로써
# 일부 데이터의 모음만을 가지고 안정적으로 weight update를 하는 방법이다.


# Decay LR by a factor of 0.1 every 7 epochs
exp_lr_scheduler = lr_scheduler.StepLR(optimizer_ft, step_size=7, gamma=0.1)


# Learning Rate Scheduler는 특정 Epoch 마다 Learning Rate를 다르게 해서 더 효율적으로 Global Minima에 수렴할 수 있게 해줌
# 예를 들면,
# 1~10 epoch : Learning Rate 0.1
# 11~20 epoch : Learning Rate 0.01
# 21~30 epoch : Learning Rate 0.001

# Learning Rate가 너무 작으면 Local Minima에 빠질 확률이 높아지고,
# Learning Rate가 너무 크면 가중치가 학습될 때 Oscilate 되는 현상이 발생해서, Global Minima에 수렴할 확률이 낮아지는 문제 방지
# -----------------------------------------------------------------------------------------------------------------------


# -----------------------------------------------------------------------------------------------------------------------
# Training Set 기반 신경망 학습 & Test Set 기반 신경망 훈련도 검사
# 각 epoch마다 Training accuracy와 Test accuracy를 확인할 수 있다.
# 즉, epoch을 크게해서 한 번 실행 하고나면, 적어도 그 epoch의 범위 내에서는 최적의 epoch 값을 찾을 수 있다는 것이다.
# -----------------------------------------------------------------------------------------------------------------------
def train_model(model, criterion, optimizer, scheduler, num_epochs=25):
    since = time.time()  # 시작 시간을 기록(총 소요 시간 계산을 위해)

    epoch = 0

    best_model_wts = copy.deepcopy(model.state_dict())
    best_acc_train = 0.0
    best_acc_validation = 0.0
    epoch_loss_train = 0.0
    epoch_acc_train = 0.0
    epoch_loss_validation = 0.0
    epoch_acc_validation = 0.0

    condition_train_end = False
    condition_val_end = False

    if USE_VISDOM:
        vis1 = visdom.Visdom()
        vis1.close()
        vis2 = visdom.Visdom()
        vis2.close()
        vis3 = visdom.Visdom()
        vis3.close()
        vis4 = visdom.Visdom()
        vis4.close()

        plot1 = vis1.line(Y=[0], X=[0], opts=dict(title='Train: Loss', showlegend=False,
                                                  xtickmin=0, ytickmin=0, ytickmax=2, ytickstep=0.5))
        plot2 = vis2.line(Y=[0], X=[0], opts=dict(title='Train: Accuracy', showlegend=False,
                                                  xtickmin=0, ytickmin=0, ytickmax=1, ytickstep=0.5))
        plot3 = vis3.line(Y=[0], X=[0], opts=dict(title='Validation: Loss', showlegend=False,
                                                  xtickmin=0, ytickmin=0, ytickmax=2, ytickstep=0.5))
        plot4 = vis4.line(Y=[0], X=[0], opts=dict(title='Validation: Accuracy', showlegend=False,
                                                  xtickmin=0, ytickmin=0, ytickmax=1, ytickstep=0.5))

        visBatchImg1 = visdom.Visdom()
        visBatchImg1.close()
        inp, classes = next(iter(dataloaders['train']))
        visBatchImg1.image(torchvision.utils.make_grid(inp), opts=dict(title="Train", caption=""))

    optimizer.step()  # Warning 제거??

    # for epoch in range(num_epochs):
    while True:
        flag_save_model = False

        # 성공 조건 만족 할 때까지 무한 루프
        if condition_train_end:
            if condition_val_end:
                print("\nNo need to do more training & validation !!!\n\n")
                break

        epoch = epoch + 1

        if epoch > NUM_OF_EPOCHS:
            print("\nTrying to train too many epoch, something wrong with network...\n\n")
            break

        print('-' * 15)
        print('Epoch {:04d}/{:04d}'.format(epoch, num_epochs))  # epoch 카운트
        print('-' * 15)

        # Each epoch has a training and validation phase
        for phase in ['train', 'val']:  # train mode -> validation mode 순으로 진행
            if phase == 'train':
                scheduler.step()
                model.train()  # Set model to training mode
            else:
                model.eval()  # Set model to evaluate mode

            running_loss = 0.0
            running_corrects = 0

            correct = 0

            # Iterate over data.
            for inputs, labels in dataloaders[phase]:  # dataloader로부터 dataset과 그에 해당되는 label을 불러옴
                inputs = inputs.to(device)  # GPU로 입력데이터를 올림
                labels = labels.to(device)  # GPU로 label을 올림

                # zero the parameter gradients
                optimizer.zero_grad()  # Gradient를 0으로 초기화

                # forward
                # track history if only in train
                with torch.set_grad_enabled(phase == 'train'):
                    outputs = model(inputs)
                    _, preds = torch.max(outputs, 1)  # 마지막 layer에서 가장 값이 큰 1개의 class를 예측 값으로 지정
                    loss = criterion(outputs, labels)

                    correct += preds.eq(labels.view_as(preds)).sum().item()

                    # backward + optimize only if in training phase
                    if phase == 'train':  # training 모드에서는 weight를 update한다.
                        loss.backward()
                        optimizer.step()

                # statistics
                running_loss += loss.item() * inputs.size(0)
                running_corrects += torch.sum(preds == labels.data)

            epoch_loss = running_loss / dataset_sizes[phase]
            epoch_acc = running_corrects.double() / dataset_sizes[phase]

            print('{} Loss: {:.4f} Acc: {:.4f} ({}/{})'.format(phase, epoch_loss, epoch_acc,
                                                               correct, len(dataloaders[phase].dataset)))

            if phase == 'train':
                epoch_loss_train = epoch_loss
                epoch_acc_train = epoch_acc

                if USE_VISDOM:
                    vis1.line(Y=[float(epoch_loss)], X=[epoch-1], win=plot1, update='append',
                              opts=dict(title='Train: Loss', showlegend=False, xtickmin=0, ytickmin=0, ytickmax=2, ytickstep=0.5))
                    vis2.line(Y=[float(epoch_acc)], X=[epoch-1], win=plot2, update='append',
                              opts=dict(title='Train: Accuracy', showlegend=False, xtickmin=0, ytickmin=0, ytickmax=1, ytickstep=0.5))

                if correct == len(dataloaders[phase].dataset):
                    if epoch_acc == 1:
                        condition_train_end = True
            else:
                epoch_loss_validation = epoch_loss
                epoch_acc_validation = epoch_acc

                if epoch_acc_validation >= best_acc_validation:
                    best_acc_validation = epoch_acc_validation
                    # deep copy the model
                    best_model_wts = copy.deepcopy(model.state_dict())

                    if epoch_acc_train > best_acc_train:
                        best_acc_train = epoch_acc_train
                        flag_save_model = True

                if USE_VISDOM:
                    vis3.line(Y=[float(epoch_loss)], X=[epoch-1], win=plot3, update='append',
                              opts=dict(title='Validation: Loss', showlegend=False, xtickmin=0, ytickmin=0, ytickmax=2, ytickstep=0.5))
                    vis4.line(Y=[float(epoch_acc)], X=[epoch-1], win=plot4, update='append',
                              opts=dict(title='Validation: Accuracy', showlegend=False, xtickmin=0, ytickmin=0, ytickmax=1, ytickstep=0.5))

                if correct == len(dataloaders[phase].dataset):
                    if epoch_acc == 1:
                        condition_val_end = True

            # Save best model weights to file
            if flag_save_model and phase == 'val':
                filename = 'MyWork_E{:04d}_Tra_A{:.4f}_L{:.4f}_Val_A{:.4f}_L{:.4f}'.format(
                               epoch, epoch_acc_train, epoch_loss_train, epoch_acc_validation, epoch_loss_validation)

                print('Save:', filename)
                torch.save(model.state_dict(), './' + filename + '_model.pth', pickle_protocol=4)
                torch.save(optimizer_ft.state_dict(), './' + filename + '_optimizer.pth', pickle_protocol=4)

                # -----------------------------------------------------------------------------------------------------------------------
                # PyTorch 모델을 ONNX로 변환하고 ONNX 런타임에서 실행
                # -----------------------------------------------------------------------------------------------------------------------
                # 모델에 대한 입력 값 생성
                #dummy_data = torch.empty(SIZE_OF_BATCH, INPUT_DIMENSION, INPUT_H, INPUT_W)
                dummy_data = torch.rand(SIZE_OF_BATCH, INPUT_DIMENSION, INPUT_H, INPUT_W)
                dummy_data = dummy_data.to(device)  # GPU 로 입력 데이터를 올림
                torch_out = model(dummy_data)

                # 모델을 추론(평가) 모드로 전환
                # 이는 dropout이나 batchnorm과 같은 연산들이 추론과 학습 모드에서 다르게 작동하기 떄문
                model.eval()   # validation 단계에서 이미 eval() 모드 이지만 명시적으로 실행

                # 모델 변환
                #torch.onnx.export(model,  # 실행될 모델
                #                  dummy_data,  # 모델 입력값 (튜플 또는 여러 입력값들도 가능)
                #                  './' + filename + '_model.onnx',  # 모델 저장 경로 (파일 또는 파일과 유사한 객체 모두 가능)
                #                  export_params=True,  # 모델 파일 안에 학습된 모델 가중치를 저장할지의 여부
                #                  opset_version=9,  # 모델을 변환할 때 사용할 ONNX 버전 (9 = ONNX v1.4)
                #                  do_constant_folding=True,  # 최적화 시 상수폴딩을 사용할지의 여부
                #                  input_names=['input'],  # 모델의 입력값을 가리키는 이름
                #                  output_names=['output'],  # 모델의 출력값을 가리키는 이름
                #                  dynamic_axes={'input': {0: 'SIZE_OF_BATCH'},  # 가변적인 길이를 가진 차원
                #                                'output': {0: 'SIZE_OF_BATCH'}})

                torch.onnx.export(model,  # 실행될 모델
                                  dummy_data,  # 모델 입력값 (튜플 또는 여러 입력값들도 가능)
                                  './' + filename + '_model.onnx',  # 모델 저장 경로 (파일 또는 파일과 유사한 객체 모두 가능)
                                  opset_version=9)  # 모델을 변환할 때 사용할 ONNX 버전 (9 = ONNX v1.4)

                # ONNX 런타임에서의 모델 결과값을 확인하기 전에 먼저 ONNX API를 사용해 ONNX 모델을 확인
                onnx_model = onnx.load('./' + filename + '_model.onnx')
                onnx.checker.check_model(onnx_model)

                # Print a human readable representation of the graph
                # print(onnx.helper.printable_graph(onnx_model.graph))

                # 이제 ONNX 런타임(ort = ONNX Run-Time)의 Python API를 통해 결과값을 계산
                # 모델이 ONNX 런타임과 PyTorch에서 동일한 결과를 출력하는지를 확인하기 위해 동일한 프로세스에서 계속 실행
                # 모델을 위한 추론 세션을 생성
                ort_session = onnxruntime.InferenceSession('./' + filename + '_model.onnx')

                def to_numpy(tensor):
                    return tensor.detach().cpu().numpy() if tensor.requires_grad else tensor.cpu().numpy()

                # 모델의 run() API를 사용하여 모델을 실행 -> ort_outs 는 ONNX 런타임에서 연산된 모델의 결과값들을 포함하고 있는 리스트
                # ONNX 런타임에서 계산된 결과값
                ort_inputs = {ort_session.get_inputs()[0].name: to_numpy(dummy_data)}
                ort_outs = ort_session.run(None, ort_inputs)

                # ONNX 런타임과 PyTorch에서 연산된 결과값 비교
                # (PyTorch와 ONNX 런타임에서 연산된 결과값이 서로 일치하는지 오차범위 (rtol=1e-03, atol=1e-05) 이내에서 확인)
                np.testing.assert_allclose(to_numpy(torch_out), ort_outs[0], rtol=1e-03, atol=1e-05)

                # ASSERTION (assert_allclose)이 발생하지 않으면 성공
                # print("Exported model has been tested with ONNXRuntime, and the result looks good!\n\n")
                # -----------------------------------------------------------------------------------------------------------------------
        print()

    time_elapsed = time.time() - since
    print('Training complete in {:.0f}m {:.0f}s'.format(time_elapsed // 60, time_elapsed % 60))

    print('Best Accuracy : Train ({:4f}), Validation ({:4f})'.format(best_acc_train, best_acc_validation))

    print()
    print()

    # load best model weights
    model.load_state_dict(best_model_wts)
    return model


# -----------------------------------------------------------------------------------------------------------------------


# 학습 시작
model_ft = train_model(model_ft, criterion, optimizer_ft, exp_lr_scheduler, NUM_OF_EPOCHS)


